from __future__ import annotations
from contextlib import contextmanager
from typing import List, Tuple, Iterator, Optional, Dict
import copy
import torch
import numpy as np
from deepchem.utils import memoize_method
from deepchem.utils.dft_utils import AtomCGTOBasis, CGTOBasis, Lattice
from deepchem.utils.dft_utils.hamilton.intor.utils import np2ctypes, int2ctypes, NDIM, CINT

# Terminology:
# * gauss: one gaussian element (multiple gaussian becomes one shell)
# * shell: one contracted basis (the same contracted gaussian for different atoms
#          counted as different shells)
# * ao: shell that has been splitted into its components,
#       e.g. p-shell is splitted into 3 components for cartesian (x, y, z)

PTR_RINV_ORIG = 4  # from libcint/src/cint_const.h


class LibcintWrapper(object):
    """ A class to wrap the libcint environment and parameters for the
    integrals calculation. This class is used to store the environment and
    parameters for the integrals calculation. It also provides the slicing
    functionality to get the subset of the shells.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import AtomCGTOBasis, LibcintWrapper, loadbasis
    >>> dtype = torch.double
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> wrap = LibcintWrapper(atombases, True, None)
    >>> wrap.ao_idxs()
    (0, 6)

    """

    def __init__(self,
                 atombases: List[AtomCGTOBasis],
                 spherical: bool = True,
                 lattice: Optional[Lattice] = None) -> None:
        """Initialize the LibcintWrapper object.

        Parameters
        ----------
        atombases: List[AtomCGTOBasis]
            List of AtomCGTOBasis objects that contains the basis set for each atom
        spherical: bool, optional (default=True)
            Whether the basis is in spherical coordinate (otherwise, it is in
            cartesian coordinate)
        lattice: Optional[Lattice], optional (default=None)
            The lattice of the system (if any)

        """
        self._atombases = atombases
        self._spherical = spherical
        self._fracz = False
        self._natoms = len(atombases)
        self._lattice = lattice

        # get dtype and device for torch's tensors
        self.dtype = atombases[0].bases[0].alphas.dtype
        self.device = atombases[0].bases[0].alphas.device

        # construct _atm, _bas, and _env as well as the parameters
        ptr_env = 20  # initial padding from libcint
        atm_list: List[List[int]] = []
        env_list: List[float] = [0.0] * ptr_env
        bas_list: List[List[int]] = []
        allpos: List[torch.Tensor] = []
        allalphas: List[torch.Tensor] = []
        allcoeffs: List[torch.Tensor] = []
        allangmoms: List[int] = []
        shell_to_atom: List[int] = []
        ngauss_at_shell: List[int] = []
        gauss_to_shell: List[int] = []

        # constructing the triplet lists and also collecting the parameters
        nshells = 0
        ishell = 0
        for iatom, atombasis in enumerate(atombases):
            # construct the atom environment
            assert atombasis.pos.numel(
            ) == NDIM, "Please report this bug in Github"
            atomz = atombasis.atomz
            #                charge    ptr_coord, nucl model (unused for standard nucl model)
            atm_list.append([int(atomz), ptr_env, 1, ptr_env + NDIM, 0, 0])
            env_list.extend(atombasis.pos.detach())
            env_list.append(0.0)
            ptr_env += NDIM + 1

            # check if the atomz is fractional
            if isinstance(atomz, float) or \
                    (isinstance(atomz, torch.Tensor) and atomz.is_floating_point()):
                self._fracz = True

            # add the atom position into the parameter list
            # TODO: consider moving allpos into shell
            allpos.append(atombasis.pos.unsqueeze(0))

            nshells += len(atombasis.bases)
            shell_to_atom.extend([iatom] * len(atombasis.bases))

            # then construct the basis
            for shell in atombasis.bases:
                assert shell.alphas.shape == shell.coeffs.shape and shell.alphas.ndim == 1, \
                    "Please report this bug in Github"
                shell.wfnormalize_()
                ngauss = len(shell.alphas)
                #                iatom, angmom,       ngauss, ncontr, kappa, ptr_exp
                bas_list.append([
                    iatom,
                    shell.angmom,
                    ngauss,
                    1,
                    0,
                    ptr_env,
                    # ptr_coeffs,           unused
                    ptr_env + ngauss,
                    0
                ])
                env_list.extend(shell.alphas.detach())
                env_list.extend(shell.coeffs.detach())
                ptr_env += 2 * ngauss

                # add the alphas and coeffs to the parameters list
                allalphas.append(shell.alphas)
                allcoeffs.append(shell.coeffs)
                allangmoms.extend([shell.angmom] * ngauss)
                ngauss_at_shell.append(ngauss)
                gauss_to_shell.extend([ishell] * ngauss)
                ishell += 1

        # compile the parameters of this object
        self._allpos_params = torch.cat(allpos, dim=0)  # (natom, NDIM)
        self._allalphas_params = torch.cat(allalphas, dim=0)  # (ntot_gauss)
        self._allcoeffs_params = torch.cat(allcoeffs, dim=0)  # (ntot_gauss)
        self._allangmoms = torch.tensor(allangmoms,
                                        dtype=torch.int32,
                                        device=self.device)  # (ntot_gauss)
        self._gauss_to_shell = torch.tensor(gauss_to_shell,
                                            dtype=torch.int32,
                                            device=self.device)

        # convert the lists to numpy to make it contiguous (Python lists are not contiguous)
        self._atm = np.array(atm_list, dtype=np.int32, order="C")
        self._bas = np.array(bas_list, dtype=np.int32, order="C")
        self._env = np.array(env_list, dtype=np.float64, order="C")

        # construct the full shell mapping
        shell_to_aoloc = [0]
        ao_to_shell: List[int] = []
        ao_to_atom: List[int] = []
        for i in range(nshells):
            nao_at_shell_i = self._nao_at_shell(i)
            shell_to_aoloc_i = shell_to_aoloc[-1] + nao_at_shell_i
            shell_to_aoloc.append(shell_to_aoloc_i)
            ao_to_shell.extend([i] * nao_at_shell_i)
            ao_to_atom.extend([shell_to_atom[i]] * nao_at_shell_i)

        self._ngauss_at_shell_list = ngauss_at_shell
        self._shell_to_aoloc = np.array(shell_to_aoloc, dtype=np.int32)
        self._shell_idxs = (0, nshells)
        self._ao_to_shell = torch.tensor(ao_to_shell,
                                         dtype=torch.long,
                                         device=self.device)
        self._ao_to_atom = torch.tensor(ao_to_atom,
                                        dtype=torch.long,
                                        device=self.device)

    @property
    def parent(self) -> LibcintWrapper:
        """parent is defined as the full LibcintWrapper where it takes the
        full shells for the integration (without the need for subsetting)

        Returns
        -------
        LibcintWrapper
            The parent of this object

        """
        return self

    @property
    def natoms(self) -> int:
        """Number of atoms in the environment

        Returns
        -------
        int
            Number of atoms in the environment

        """
        return self._natoms

    @property
    def fracz(self) -> bool:
        """Indicating whether we are working with fractional z

        Returns
        -------
        bool
            Whether we are working with fractional z

        """
        return self._fracz

    @property
    def lattice(self) -> Optional[Lattice]:
        """The lattice of the system

        Returns
        -------
        Optional[Lattice]
            The lattice of the system

        """
        return self._lattice

    @property
    def spherical(self) -> bool:
        """returns whether the basis is in spherical coordinate (otherwise, it
        is in cartesian coordinate)

        Returns
        -------
        bool
            Whether the basis is in spherical coordinate

        """
        return self._spherical

    @property
    def atombases(self) -> List[AtomCGTOBasis]:
        """List of AtomCGTOBasis objects that contains the basis set for each atom

        Returns
        -------
        List[AtomCGTOBasis]
            List of AtomCGTOBasis objects that contains the basis set for each atom

        """
        return self._atombases

    @property
    def atm_bas_env(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """returns the triplet lists, i.e. atm, bas, env
        this shouldn't change in the sliced wrapper

        Returns
        -------
        Tuple[np.ndarray, np.ndarray, np.ndarray]
            The triplet lists, i.e. atm, bas, env

        """
        return self._atm, self._bas, self._env

    @property
    def full_angmoms(self) -> torch.Tensor:
        """Returns the full angular momentums tensor.

        Returns
        -------
        torch.Tensor
            The full angular momentums tensor.

        """
        return self._allangmoms

    @property
    def params(self) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Returns all the parameters of this object.

        This shouldn't change in the sliced wrapper.

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
            All the parameters of this object.

        """

        return self._allcoeffs_params, self._allalphas_params, self._allpos_params

    @property
    def shell_idxs(self) -> Tuple[int, int]:
        """Returns the lower and upper indices of the shells of this object
        in the absolute index (upper is exclusive).

        Returns
        -------
        Tuple[int, int]
            The lower and upper indices of the shells.
        """
        return self._shell_idxs

    @property
    def full_shell_to_aoloc(self) -> np.ndarray:
        """Returns the full array mapping from shell index to absolute AO location.

        Returns
        -------
        np.ndarray
            The full array mapping from shell index to absolute AO location.
        """
        return self._shell_to_aoloc

    @property
    def full_gauss_to_shell(self) -> torch.Tensor:
        """Returns the full index mapping from gaussian to shell tensor.

        Returns
        -------
        torch.Tensor
            The full index mapping from gaussian to shell tensor.
        """
        return self._gauss_to_shell

    @property
    def full_ao_to_atom(self) -> torch.Tensor:
        """Returns the full array mapping from atomic orbital index to the atom location.

        Returns
        -------
        torch.Tensor
            The full array mapping from atomic orbital index to the atom location.
        """
        return self._ao_to_atom

    @property
    def full_ao_to_shell(self) -> torch.Tensor:
        """Returns the full array mapping from atomic orbital index to the shell location.

        Returns
        -------
        torch.Tensor
            The full array mapping from atomic orbital index to the shell location.
        """
        return self._ao_to_shell

    @property
    def ngauss_at_shell(self) -> List[int]:
        """Returns the number of gaussian basis at the given shell.

        Returns
        -------
        List[int]
            The number of gaussian basis at the given shell.
        """
        return self._ngauss_at_shell_list

    @memoize_method
    def __len__(self) -> int:
        """Returns the total number of shells.

        Returns
        -------
        int
            The total number of shells.
        """
        return self.shell_idxs[-1] - self.shell_idxs[0]

    @memoize_method
    def nao(self) -> int:
        """Returns the number of atomic orbitals.

        Returns
        -------
        int
            The number of atomic orbitals.
        """
        shell_idxs = self.shell_idxs
        return self.full_shell_to_aoloc[
            shell_idxs[-1]] - self.full_shell_to_aoloc[shell_idxs[0]]

    @memoize_method
    def ao_idxs(self) -> Tuple[int, int]:
        """Returns the lower and upper indices of the atomic orbitals.

        Returns
        -------
        Tuple[int, int]
            The lower and upper indices of the atomic orbitals.
        """
        shell_idxs = self.shell_idxs
        return self.full_shell_to_aoloc[
            shell_idxs[0]], self.full_shell_to_aoloc[shell_idxs[1]]

    @memoize_method
    def ao_to_atom(self) -> torch.Tensor:
        """Returns the relative mapping from atomic orbital relative index to the absolute atom position.

        Returns
        -------
        torch.Tensor
            The relative mapping from atomic orbital relative index to the absolute atom position.
        """
        return self.full_ao_to_atom[slice(*self.ao_idxs())]

    @memoize_method
    def ao_to_shell(self) -> torch.Tensor:
        """Returns the relative mapping from atomic orbital relative index to the absolute shell position.

        Returns
        -------
        torch.Tensor
            The relative mapping from atomic orbital relative index to the absolute shell position.
        """
        return self.full_ao_to_shell[slice(*self.ao_idxs())]

    def __getitem__(self, inp) -> LibcintWrapper:
        """Gets the subset of the shells, but keeping the environment and parameters the same.

        Returns
        -------
        LibcintWrapper
            The subset of the shells with the same environment and parameters.
        """
        assert isinstance(inp, slice)
        assert inp.step is None or inp.step == 1
        assert inp.start is not None or inp.stop is not None

        # complete the slice
        nshells = self.shell_idxs[-1]
        if inp.start is None and inp.stop is not None:
            inp = slice(0, inp.stop)
        elif inp.start is not None and inp.stop is None:
            inp = slice(inp.start, nshells)

        # make the slice positive
        if inp.start < 0:
            inp = slice(inp.start + nshells, inp.stop)
        if inp.stop < 0:
            inp = slice(inp.start, inp.stop + nshells)

        return SubsetLibcintWrapper(self, inp)

    @memoize_method
    def get_uncontracted_wrapper(self) -> Tuple[LibcintWrapper, torch.Tensor]:
        """Returns the uncontracted LibcintWrapper and the mapping from uncontracted to contracted AO indices.

        Returns
        -------
        Tuple[LibcintWrapper, torch.Tensor]
            The uncontracted LibcintWrapper and the mapping from uncontracted to contracted AO indices.
        """
        new_atombases = []
        for atombasis in self.atombases:
            atomz = atombasis.atomz
            pos = atombasis.pos
            new_bases = []
            for shell in atombasis.bases:
                angmom = shell.angmom
                alphas = shell.alphas
                coeffs = shell.coeffs
                normalized = shell.normalized
                new_bases.extend([
                    CGTOBasis(angmom,
                              alpha[None],
                              coeff[None],
                              normalized=normalized)
                    for (alpha, coeff) in zip(alphas, coeffs)
                ])
            new_atombases.append(
                AtomCGTOBasis(atomz=atomz, bases=new_bases, pos=pos))
        uncontr_wrapper = LibcintWrapper(new_atombases,
                                         spherical=self.spherical)

        # get the mapping uncontracted ao to the contracted ao
        uao2ao: List[int] = []
        idx_ao = 0
        # iterate over shells
        for i in range(len(self)):
            nao = self._nao_at_shell(i)
            uao2ao += list(range(idx_ao,
                                 idx_ao + nao)) * self.ngauss_at_shell[i]
            idx_ao += nao
        uao2ao_res = torch.tensor(uao2ao, dtype=torch.long, device=self.device)
        return uncontr_wrapper, uao2ao_res

    @staticmethod
    def concatenate(*wrappers: LibcintWrapper) \
            -> Tuple[LibcintWrapper, ...]:
        """
        Concatenate the parents of wrappers, then returns the subsets
        corresponds to the wrappers.
        This function returns an environment that contains all the atoms from
        the parents of all the wrappers.
        If all the wrappers are from the same parent, then this function does
        not do anything.

        Parameters
        ----------
        wrappers : Tuple[LibcintWrapper, ...]
            The wrappers to concatenate.

        Returns
        -------
        Tuple[LibcintWrapper, ...]
            The subsets corresponding to the input wrappers.

        """

        # construct the parent mapping
        unique_parents: List[LibcintWrapper] = []
        unique_pids: Dict[int, int] = {}
        w2pidx: List[int] = []
        shell_idxs: List[Tuple[int, int]] = []
        cumsum_plen: List[int] = [0]
        for w in wrappers:
            parent = w.parent
            pid = id(parent)
            shell_idxs.append(w.shell_idxs)
            if pid not in unique_pids:
                idx = len(unique_parents)
                unique_pids[pid] = idx
                unique_parents.append(parent)
                cumsum_plen.append(cumsum_plen[-1] + len(parent))
            w2pidx.append(unique_pids[pid])

        # check the length of the unique parents, if there is only 1 unique parent,
        # then just return as it is
        assert len(unique_parents) > 0
        if len(unique_parents) == 1:
            return (*wrappers,)

        # check if the unique parents have the same options (except for the atombases)
        p0 = unique_parents[0]
        sph = p0.spherical
        latt = p0.lattice
        atombases = copy.copy(p0.atombases)  # shallow copy
        for p in unique_parents[1:]:
            assert p.spherical == sph
            assert p.lattice == latt
            atombases.extend(p.atombases)

        # get the grand environment
        grandp = LibcintWrapper(atombases, spherical=sph, lattice=latt)

        # get the subsets which correspond to the input wrappers
        res: List[LibcintWrapper] = []
        for i, w in enumerate(wrappers):
            sh_idx0, sh_idx1 = shell_idxs[i]
            offset = cumsum_plen[w2pidx[i]]
            sh_idx0 = sh_idx0 + offset
            sh_idx1 = sh_idx1 + offset
            res.append(grandp[sh_idx0:sh_idx1])

        return (*res,)

    # misc functions
    @contextmanager
    def centre_on_r(self, r: torch.Tensor) -> Iterator:
        """Sets the centre of coordinate to r.

        Parameters
        ----------
        r : torch.Tensor
            The coordinates to set the centre to.

        Yields
        ------
        Iterator
            A context manager.
        """
        try:
            env = self.atm_bas_env[-1]
            prev_centre = env[PTR_RINV_ORIG:PTR_RINV_ORIG + NDIM]
            env[PTR_RINV_ORIG:PTR_RINV_ORIG + NDIM] = r.detach().numpy()
            yield
        finally:
            env[PTR_RINV_ORIG:PTR_RINV_ORIG + NDIM] = prev_centre

    def _nao_at_shell(self, sh: int) -> int:
        """Returns the number of atomic orbital at the given shell index.

        Parameters
        ----------
        sh : int
            The shell index.

        Returns
        -------
        int
            The number of atomic orbital at the given shell index.
        """
        if self.spherical:
            op = CINT().CINTcgto_spheric
        else:
            op = CINT().CINTcgto_cart
        bas = self.atm_bas_env[1]
        return op(int2ctypes(sh), np2ctypes(bas))


class SubsetLibcintWrapper(LibcintWrapper):
    """
    A class to represent the subset of LibcintWrapper.
    If put into integrals or evaluations, this class will only evaluate
        the subset of the shells from its parent.
    The environment will still be the same as its parent.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import AtomCGTOBasis, LibcintWrapper, loadbasis
    >>> dtype = torch.double
    >>> d = 1.0
    >>> pos_requires_grad = True
    >>> pos1 = torch.tensor([0.1 * d,  0.0 * d,  0.2 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos2 = torch.tensor([0.0 * d,  1.0 * d, -0.4 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> pos3 = torch.tensor([0.2 * d, -1.4 * d, -0.9 * d], dtype=dtype, requires_grad=pos_requires_grad)
    >>> poss = [pos1, pos2, pos3]
    >>> atomzs = [1, 1, 1]
    >>> allbases = [
    ...     loadbasis("%d:%s" % (max(atomz, 1), "3-21G"), dtype=dtype, requires_grad=False)
    ...     for atomz in atomzs
    ... ]
    >>> atombases = [
    ...     AtomCGTOBasis(atomz=atomzs[i], bases=allbases[i], pos=poss[i])
    ...     for i in range(len(allbases))
    ... ]
    >>> wrap = LibcintWrapper(atombases, True, None)
    >>> subset = wrap[1:3]
    >>> subset.ao_idxs()
    (1, 3)

    """

    def __init__(self, parent: LibcintWrapper, subset: slice):
        """
        Initializes the SubsetLibcintWrapper.

        Parameters
        ----------
        parent : LibcintWrapper
            The parent LibcintWrapper.
        subset : slice
            The slice representing the subset of shells.
        """
        self._parent = parent
        self._shell_idxs = subset.start, subset.stop

    @property
    def parent(self) -> LibcintWrapper:
        """
        Returns the parent LibcintWrapper.

        Returns
        -------
        LibcintWrapper
            The parent LibcintWrapper.
        """
        return self._parent

    @property
    def shell_idxs(self) -> Tuple[int, int]:
        """
        Returns the shell indices of the subset.

        Returns
        -------
        Tuple[int, int]
            The shell indices of the subset.
        """
        return self._shell_idxs

    @memoize_method
    def get_uncontracted_wrapper(self):
        """
        Returns the uncontracted LibcintWrapper and the mapping from
        uncontracted atomic orbital (relative index) to the relative index
        of the atomic orbital of the contracted wrapper.

        Returns
        -------
        Tuple[LibcintWrapper, torch.Tensor]
            The uncontracted LibcintWrapper and the mapping from uncontracted
            atomic orbital (relative index) to the relative index of the atomic
            orbital of the contracted wrapper.
        """
        pu_wrapper, p_uao2ao = self._parent.get_uncontracted_wrapper()

        # determine the corresponding shell indices in the new uncontracted wrapper
        shell_idxs = self.shell_idxs
        gauss_idx0 = sum(self._parent.ngauss_at_shell[:shell_idxs[0]])
        gauss_idx1 = sum(self._parent.ngauss_at_shell[:shell_idxs[1]])
        u_wrapper = pu_wrapper[gauss_idx0:gauss_idx1]

        # construct the uao (relative index) mapping to the absolute index
        # of the atomic orbital in the contracted basis
        uao2ao = []
        idx_ao = 0
        for i in range(shell_idxs[0], shell_idxs[1]):
            nao = self._parent._nao_at_shell(i)
            uao2ao += list(range(
                idx_ao, idx_ao + nao)) * self._parent.ngauss_at_shell[i]
            idx_ao += nao
        uao2ao_res = torch.tensor(uao2ao, dtype=torch.long, device=self.device)
        return u_wrapper, uao2ao_res

    def __getitem__(self, inp):
        """
        Raises NotImplementedError as indexing of SubsetLibcintWrapper is not implemented.

        Parameters
        ----------
        inp : Any
            The input for indexing.

        Raises
        ------
        NotImplementedError
            Indexing of SubsetLibcintWrapper is not implemented.
        """
        raise NotImplementedError(
            "Indexing of SubsetLibcintWrapper is not implemented")

    def __getattr__(self, name):
        """
        Delegate attribute access to the parent LibcintWrapper.

        Parameters
        ----------
        name : str
            The name of the attribute.

        Returns
        -------
        Any
            The attribute value from the parent LibcintWrapper.
        """
        return getattr(self._parent, name)



import math
import deepchem as dc
import torch
import torch.nn as nn
from torch_geometric.data import Batch
from torch_geometric.nn import global_add_pool, radius
from torch_geometric.utils import remove_self_loops
from typing import Union, Callable, Tuple, Optional
from deepchem.utils.pytorch_utils import get_activation
from deepchem.models.torch_models.layers import MXMNetBesselBasisLayer, MXMNetSphericalBasisLayer, MultilayerPerceptron


class MXMNet(nn.Module):
    """
    Molecular Mechanics-Driven Graph Neural Network with Multiplex Graph for Molecular Structures.
    In this class, we define the global and local message passing layers and the MXMNet Model[1]_.
    We also define the forward call of this model in the forward function.

    Note:
    This model currently supports only Single Task Regression. Multi Task Regression, Single-Task Classification
    and Multi-Task Classification is not yet implemented.
    References
    ----------
    .. [1] Molecular Mechanics-Driven Graph Neural Network with Multiplex Graph for Molecular Structures https://arxiv.org/pdf/2011.07457

    Example
    -------
    >>> import deepchem as dc
    >>> import os
    >>> import tempfile
    >>> import torch
    >>> from torch_geometric.data import Data, Batch
    >>> from deepchem.feat.molecule_featurizers import MXMNetFeaturizer
    >>> QM9_TASKS = ["mu", "alpha", "homo", "lumo", "gap", "r2", "zpve", "cv", "u0", "u298",
    ...              "h298", "g298"]
    >>> dim = 10
    >>> n_layer = 6
    >>> cutoff = 5
    >>> feat = MXMNetFeaturizer()
    >>> # Get data
    >>> loader = dc.data.SDFLoader(tasks=[QM9_TASKS[0]],
    ...                            featurizer=feat,
    ...                            sanitize=True)
    >>> current_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    >>> dataset_path = os.path.join(current_dir, "tests/assets/qm9_mini.sdf")
    >>> dataset = loader.create_dataset(inputs=dataset_path,
    ...                            shard_size=1)
    >>> model = MXMNet(dim=dim, n_layer=n_layer, cutoff=cutoff)
    >>> train_dir = None
    >>> if train_dir is None:
    ...    train_dir = tempfile.mkdtemp()
    >>> data = dataset.select([i for i in range(0, 1)], train_dir)
    >>> data = data.X
    >>> data = [data[i].to_pyg_graph() for i in range(1)]
    >>> pyg_batch = Batch()
    >>> pyg_batch = pyg_batch.from_data_list(data)
    >>> output = model(pyg_batch)
    >>> output[0].shape
    torch.Size([1])
    """

    def __init__(self,
                 dim: int,
                 n_layer: int,
                 cutoff: float,
                 num_spherical: int = 7,
                 num_radial: int = 6,
                 envelope_exponent: int = 5,
                 activation_fn: Union[Callable, str] = 'silu',
                 n_tasks: int = 1):
        """Initialize the MXMNet class.
        Parameters
        ----------
        dim: int
            The dimensionality of node embeddings.
        n_layer: int
            The number of message passing layers.
        cutoff: float
            The distance cutoff for edge connections.
        num_spherical: int, default 7
            The number of spherical harmonics.
        num_radial: int, default 6
            The number of radial basis functions.
        envelope_exponent: int, default 5
            The exponent for the envelope function.
        activation_fn: Union[Callable, str], default 'silu'
            The activation function name.
        n_tasks: int, default 1
            The number of prediction tasks. Only single Task regression is supported currently.
        """
        super(MXMNet, self).__init__()

        self.dim: int = dim
        self.n_tasks: int = n_tasks
        self.n_layer: int = n_layer
        self.cutoff: float = cutoff
        activation_fn = get_activation(activation_fn)

        self.embeddings: nn.Parameter = nn.Parameter(torch.ones((5, self.dim)))

        self.rbf_l: MXMNetBesselBasisLayer = dc.models.torch_models.layers.MXMNetBesselBasisLayer(
            16, 5, envelope_exponent)
        self.rbf_g: MXMNetBesselBasisLayer = dc.models.torch_models.layers.MXMNetBesselBasisLayer(
            16, self.cutoff, envelope_exponent)
        self.sbf: MXMNetSphericalBasisLayer = dc.models.torch_models.layers.MXMNetSphericalBasisLayer(
            num_spherical, num_radial, 5, envelope_exponent)

        self.rbf_g_mlp: MultilayerPerceptron = dc.models.torch_models.layers.MultilayerPerceptron(
            d_input=16, d_output=self.dim, activation_fn=activation_fn)
        self.rbf_l_mlp: MultilayerPerceptron = dc.models.torch_models.layers.MultilayerPerceptron(
            d_input=16, d_output=self.dim, activation_fn=activation_fn)

        self.sbf_1_mlp: MultilayerPerceptron = dc.models.torch_models.layers.MultilayerPerceptron(
            d_input=num_spherical * num_radial,
            d_output=self.dim,
            activation_fn=activation_fn)
        self.sbf_2_mlp: MultilayerPerceptron = dc.models.torch_models.layers.MultilayerPerceptron(
            d_input=num_spherical * num_radial,
            d_output=self.dim,
            activation_fn=activation_fn)

        self.global_layers: nn.ModuleList = torch.nn.ModuleList()
        for layer in range(self.n_layer):
            self.global_layers.append(
                dc.models.torch_models.layers.MXMNetGlobalMessagePassing(
                    self.dim))

        self.local_layers: nn.ModuleList = torch.nn.ModuleList()
        for layer in range(self.n_layer):
            self.local_layers.append(
                dc.models.torch_models.layers.MXMNetLocalMessagePassing(
                    self.dim))

        self.init()

    def init(self) -> None:
        """
        Initialize the node embeddings by setting them to random values within a predefined range.
        This method ensures that the node embeddings are initialized with random values, promoting diversity in the
        initial node representations.
        """

        stdv: float = math.sqrt(3)
        self.embeddings.data.uniform_(-stdv, stdv)

    def indices(
        self, edge_index: torch.Tensor, num_nodes: int
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor,
               torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor,
               torch.Tensor, torch.Tensor]:
        """
        Compute node indices for defining angles in the molecular graph.
        Parameters
        ----------
        edge_index: torch.Tensor
            The edge index tensor.
        num_nodes: int
            The number of nodes in the graph.
        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]
            Node indices for various angle calculations.
        """

        row: torch.Tensor
        col: torch.Tensor
        row, col = edge_index

        # Initialize an adjacency matrix with zeros
        adj_matrix: torch.Tensor = torch.zeros((num_nodes, num_nodes),
                                               device=row.device)

        value: torch.Tensor = torch.arange(row.size(0), device=row.device)

        for i in range(row.size(0)):
            adj_matrix[col[i], row[i]] = value[i]

        adj_t_row: torch.Tensor = adj_matrix[row]
        adj_t_row1: torch.Tensor = adj_matrix[row]

        for i in range(adj_t_row1.size(0)):
            if (adj_t_row1[i] != 0).sum() == 0:
                adj_t_row1[i][0] = 1

        # Compute the node indices for two-hop angles
        num_triplets: torch.Tensor = torch.zeros((adj_t_row.size(0)),
                                                 device=row.device)

        for i in range(adj_t_row1.size(0)):
            num_triplets[i] = (adj_t_row1[i] != 0).sum().to(torch.long)

        num_triplets = num_triplets.to(torch.long)

        idx_i: torch.Tensor = col.repeat_interleave(num_triplets)
        idx_j: torch.Tensor = row.repeat_interleave(num_triplets)
        idx_k: torch.Tensor = (adj_t_row1 != 0).nonzero()[:, 1]
        mask: torch.Tensor = idx_i != idx_k

        idx_i_1: torch.Tensor
        idx_i_1, idx_j, idx_k = idx_i[mask], idx_j[mask], idx_k[mask]
        idx_kj: torch.Tensor = adj_t_row[adj_t_row1 != 0].to(torch.long)[mask]
        idx_ji_1: torch.Tensor = (adj_t_row1 != 0).nonzero()[:, 0][mask]

        # Compute the node indices for one-hop angles
        adj_t_col: torch.Tensor = adj_matrix[col]
        adj_t_col1: torch.Tensor = adj_matrix[col]
        for i in range(adj_t_col1.size(0)):
            if (adj_t_col1[i] != 0).sum() == 0:
                adj_t_col1[i][0] = 1

        num_pairs: torch.Tensor = torch.zeros((adj_t_col.size(0)),
                                              device=row.device)
        for i in range(adj_t_col1.size(0)):
            num_pairs[i] = (adj_t_col1[i] != 0).sum().to(torch.long)

        num_pairs = num_pairs.to(torch.long)
        idx_i_2: torch.Tensor = row.repeat_interleave(num_pairs)
        idx_j1: torch.Tensor = col.repeat_interleave(num_pairs)
        idx_j2: torch.Tensor = (adj_t_col1 != 0).nonzero()[:, 1]

        idx_ji_2: torch.Tensor = (adj_t_col1 != 0).nonzero()[:, 0]
        idx_jj: torch.Tensor = adj_t_col[adj_t_col1 != 0].to(torch.long)

        return idx_i_1, idx_j, idx_k, idx_kj, idx_ji_1, idx_i_2, idx_j1, idx_j2, idx_jj, idx_ji_2

    def forward(self, pyg_batch: Batch) -> torch.Tensor:
        """
        Forward pass of the MXMNet model.
        Parameters
        ----------
        pyg_batch: Batch
            A pytorch-geometric batch containing tensors for:
                - node features
                - edge_index
                - pos
                - batch information
        Returns
        -------
        torch.Tensor
            The model's output.
        """

        x: torch.Tensor = pyg_batch.x
        edge_index: torch.Tensor = pyg_batch.edge_index
        pos: torch.Tensor = pyg_batch.pos
        batch = pyg_batch.batch
        x = torch.reshape(x, (-1,))
        # Initialize node embeddings
        h: torch.Tensor = torch.index_select(self.embeddings, 0, x.long())

        # Get the edges and pairwise distances in the local layer
        edge_index_l: torch.Tensor
        edge_index_l, _ = remove_self_loops(edge_index)
        j_l: torch.Tensor
        i_l: torch.Tensor
        j_l, i_l = edge_index_l
        dist_l: torch.Tensor = (pos[i_l] - pos[j_l]).pow(2).sum(dim=-1).sqrt()

        # Get the edges pairwise distances in the global layer
        row: torch.Tensor
        col: torch.Tensor
        row, col = radius(pos,
                          pos,
                          self.cutoff,
                          batch,
                          batch,
                          max_num_neighbors=500)
        edge_index_g: torch.Tensor = torch.stack([row, col], dim=0)
        edge_index_g, _ = remove_self_loops(edge_index_g)
        j_g: torch.Tensor
        i_g: torch.Tensor
        j_g, i_g = edge_index_g
        dist_g: torch.Tensor = (pos[i_g] - pos[j_g]).pow(2).sum(dim=-1).sqrt()

        num_nodes = h.size(0)
        # Compute the node indices for defining the angles
        idx_i_1, idx_j, idx_k, idx_kj, idx_ji, idx_i_2, idx_j1, idx_j2, idx_jj, idx_ji_2 = self.indices(
            edge_index_l, num_nodes=num_nodes)

        # Compute the two-hop angles
        pos_ji_1, pos_kj = pos[idx_j] - pos[idx_i_1], pos[idx_k] - pos[idx_j]
        a = (pos_ji_1 * pos_kj).sum(dim=-1)
        b: torch.Tensor = torch.cross(pos_ji_1, pos_kj).norm(dim=-1)
        angle_1: torch.Tensor = torch.atan2(b, a)

        # Compute the one-hop angles
        pos_ji_2, pos_jj = pos[idx_j1] - pos[idx_i_2], pos[idx_j2] - pos[idx_j1]
        a = (pos_ji_2 * pos_jj).sum(dim=-1)
        b = torch.cross(pos_ji_2, pos_jj).norm(dim=-1)
        angle_2: torch.Tensor = torch.atan2(b, a)

        # Get the RBF and SBF embeddings
        rbf_g: torch.Tensor = self.rbf_g(dist_g)
        rbf_l: torch.Tensor = self.rbf_l(dist_l)
        sbf_1: torch.Tensor = self.sbf(dist_l, angle_1, idx_kj)
        sbf_2: torch.Tensor = self.sbf(dist_l, angle_2, idx_jj)

        rbf_g = self.rbf_g_mlp(rbf_g)
        rbf_l = self.rbf_l_mlp(rbf_l)
        sbf_1 = self.sbf_1_mlp(sbf_1)
        sbf_2 = self.sbf_2_mlp(sbf_2)

        # Perform the message passing schemes
        node_sum: Optional[torch.Tensor] = None

        for layer in range(self.n_layer):
            h = self.global_layers[layer](h, rbf_g, edge_index_g)
            t: torch.Tensor
            h, t = self.local_layers[layer](h, rbf_l, sbf_1, sbf_2, idx_kj,
                                            idx_ji, idx_jj, idx_ji_2,
                                            edge_index_l)
            if (node_sum is None):
                node_sum = t
            else:
                node_sum += t

        # Readout
        output: torch.Tensor = global_add_pool(node_sum, batch)
        return output.view(-1, self.n_tasks)



import pytest
from deepchem.utils.poly_converters import PSMILES2WDGConverter, WDG2PSMILESConverter

# Sample data for testing
psmiles = "*CCCC*"
metadata = {
    "indicies": [2, 3],
    "seq_index": 4,
    "residue": "0.5|0.5|",
    "smiles_type": "SMARTS"
}
wd_graph_string = "[*:1]CC[*:2].[*:3]CC[*:4]|0.5|0.5|<1-3:0.5:0.5<1-4:0.5:0.5<2-3:0.5:0.5<2-4:0.5:0.5"


# Test PSMILES2WDGConverter class
class TestPSMILES2WDGConverter:

    def test_index_wildcards(self):
        converter = PSMILES2WDGConverter()
        result = converter.index_wildcards("*CC(*)C")
        assert result == "[1*]CC([2*])C"

    def test_add_indicies_to_smiles_from_meta(self):
        converter = PSMILES2WDGConverter()
        result = converter.add_indicies_to_smiles_from_meta("*CCCC*", 3)
        assert result == "*CC**CC*"

    def test_make_wdgraph_string_from_meta(self):
        converter = PSMILES2WDGConverter()
        result = converter.make_wdgraph_string_from_meta("*CC**CC*", [3, 4])
        assert "." in result  # Check if the bond was broken

    def test_convert_smiles_to_SMARTS(self):
        converter = PSMILES2WDGConverter()
        result = converter.convert_smiles_to_SMARTS("C*CC*")
        assert result == "C[*:1]CC[*:2]"

    def test_compose_from_meta(self):
        converter = PSMILES2WDGConverter()
        result = converter.compose_from_meta(psmiles, metadata)
        assert "|" in result  # Check if the final composed result includes metadata

    def test_convert(self):
        converter = PSMILES2WDGConverter()
        result = converter.convert(psmiles, metadata)
        assert isinstance(result, list)
        assert len(result) > 0

    def test_call(self):
        converter = PSMILES2WDGConverter()
        result = converter([psmiles], [metadata])
        assert isinstance(result, list)
        assert len(result) > 0


# Test WDG2PSMILESConverter class
class TestWDG2PSMILESConverter:

    def test_get_wildcard_bond_indecies(self):
        converter = WDG2PSMILESConverter()
        with pytest.raises(ValueError):
            converter.get_wildcard_bond_indecies("CCO")

    def test_convert_smiles_part(self):
        converter = WDG2PSMILESConverter()
        result = converter.convert_smiles_part("*CC*.*CC*")
        assert isinstance(result, tuple)
        assert len(result) == 4  # Check if the method returns 4 values

    def test_replace_SMARTS(self):
        converter = WDG2PSMILESConverter()
        result = converter.replace_SMARTS("[*:1]CC[*:2]")
        assert result == "*CC*"

    def test_convert(self):
        converter = WDG2PSMILESConverter()
        result, metadata = converter.convert(wd_graph_string)
        assert isinstance(result, str)
        assert isinstance(metadata, dict)

    def test_call(self):
        converter = WDG2PSMILESConverter()
        result, metadata = converter([wd_graph_string])
        assert isinstance(result, list)
        assert len(result) > 0



from logging import raiseExceptions
import os
import subprocess


def system_call(command):
    """ Wrapper for system command call """
    p = subprocess.Popen([command], stdout=subprocess.PIPE, shell=True)
    return p.stdout.read()


def hhblits(dataset_path,
            database=None,
            data_dir=None,
            evalue=0.001,
            num_iterations=2,
            num_threads=4):
    """
    Run hhblits multisequence alignment search on a dataset. This function
    requires the hhblits binary to be installed and in the path. This function
    also requires a Hidden Markov Model reference database to be provided. Both can be
    found here: https://github.com/soedinglab/hh-suite

    The database should be in the deepchem data directory or specified as an argument.
    To set the deepchem data directory, run this command in your environment:

    export DEEPCHEM_DATA_DIR=<path to data directory>

    Parameters
    ----------
    dataset_path: str
        Path to single sequence or multiple sequence alignment (MSA) dataset. Results will be saved in this directory.
    database: str
        Name of database to search against. Note this is not the path, but the name of the database.
    data_dir: str
        Path to database directory.
    evalue: float
        E-value cutoff.
    num_iterations: int
        Number of iterations.
    num_threads: int
        Number of threads.

    Returns
    -------
    results: .a3m file
        MSA file containing the results of the hhblits search.
    results: .hhr file
        hhsuite results file containing the results of the hhblits search.

    Examples
    --------
    >>> from deepchem.utils.sequence_utils import hhblits
    >>> msa_path = hhblits('test/data/example.fasta', database='example_db', data_dir='test/data/', evalue=0.001, num_iterations=2, num_threads=4)

    """

    if data_dir is None:
        data_dir = os.environ['DEEPCHEM_DATA_DIR']
    if len(data_dir) == 0:
        raiseExceptions(
            'hhblits requires a database. Please follow the instructions here \
    to download a database: https://github.com/soedinglab/hh-suite/wiki#hh-suite-databases'
        )

    _, dataset_file_type = os.path.splitext(dataset_path)

    save_dir = os.path.dirname(os.path.realpath(dataset_path))

    if dataset_file_type == '.fas' or '.fasta':
        command = 'hhsearch ' + \
            ' -i ' + os.path.abspath(dataset_path) + \
            ' -d ' + os.path.join(data_dir, database) + \
            ' -oa3m ' + os.path.join(save_dir, 'results.a3m') + \
            ' -cpu ' + str(num_threads) + \
            ' -n ' + str(num_iterations) + \
            ' -e ' + str(evalue) + \
            ' -M first'
    if dataset_file_type == '.a3m' or '.a2m' or '.hmm':
        command = 'hhsearch ' + \
            ' -i ' + os.path.abspath(dataset_path) + \
            ' -d ' + os.path.join(data_dir, database) + \
            ' -oa3m ' + os.path.join(save_dir, 'results.a3m') + \
            ' -cpu ' + str(num_threads) + \
            ' -n ' + str(num_iterations) + \
            ' -e ' + str(evalue)
    else:
        raiseExceptions('Unsupported file type')

    system_call(command)

    msa_path = os.path.join(save_dir, 'results.a3m')

    return msa_path


def hhsearch(dataset_path,
             database=None,
             data_dir=None,
             evalue=0.001,
             num_iterations=2,
             num_threads=4):
    """
    Run hhsearch multisequence alignment search on a dataset. This function
    requires the hhblits binary to be installed and in the path. This function
    also requires a Hidden Markov Model reference database to be provided. Both can be
    found here: https://github.com/soedinglab/hh-suite

    The database should be in the deepchem data directory or specified as an argument.
    To set the deepchem data directory, run this command in your environment:

    export DEEPCHEM_DATA_DIR=<path to data directory>

    Examples
    --------
    >>> from deepchem.utils.sequence_utils import hhsearch
    >>> msa_path = hhsearch('test/data/example.fasta', database='example_db', data_dir='test/data/', evalue=0.001, num_iterations=2, num_threads=4)

    Parameters
    ----------
    dataset_path: str
        Path to multiple sequence alignment dataset. Results will be saved in this directory.
    database: str
        Name of database to search against. Note this is not the path, but the name of the database.
    data_dir: str
        Path to database directory.
    evalue: float
        E-value cutoff.
    num_iterations: int
        Number of iterations.
    num_threads: int
        Number of threads.

    Returns
    -------
    results: .a3m file
        MSA file containing the results of the hhblits search.
    results: .hhr file
        hhsuite results file containing the results of the hhblits search.

    """

    if data_dir is None:
        data_dir = os.environ['DEEPCHEM_DATA_DIR']
    if len(data_dir) == 0:
        raiseExceptions(
            'hhsearch requires a database. Please follow the instructions here \
            to download a database: https://github.com/soedinglab/hh-suite/wiki#hh-suite-databases'
        )

    _, dataset_file_type = os.path.splitext(dataset_path)

    save_dir = os.path.dirname(os.path.abspath(dataset_path))

    if dataset_file_type == '.fas' or '.fasta':
        command = 'hhsearch ' + \
            ' -i ' + os.path.abspath(dataset_path) + \
            ' -d ' + os.path.join(data_dir, database) + \
            ' -oa3m ' + os.path.join(save_dir, 'results.a3m') + \
            ' -cpu ' + str(num_threads) + \
            ' -e ' + str(evalue) + \
            ' -M first'
    if dataset_file_type == '.a3m' or '.a2m' or '.hmm':
        command = 'hhsearch ' + \
            ' -i ' + os.path.abspath(dataset_path) + \
            ' -d ' + os.path.join(data_dir, database) + \
            ' -oa3m ' + os.path.join(save_dir, 'results.a3m') + \
            ' -cpu ' + str(num_threads) + \
            ' -e ' + str(evalue)
    else:
        raiseExceptions('Unsupported file type')

    system_call(command)

    msa_path = os.path.join(save_dir, 'results.a3m')

    return msa_path


def MSA_to_dataset(msa_path):
    """
    Convert a multiple sequence alignment to a NumpyDataset object.
    """

    from deepchem.data.datasets import NumpyDataset  # NumpyDataset depends on utils, so imported here to prevent circular import
    from Bio import SeqIO

    with open(msa_path, 'r') as f:
        ids = []
        sequences = []
        for record in SeqIO.parse(f, 'fasta'):
            ids.append(record.id)
            seq = []
            for res in record:
                seq.append(res)
            sequences.append(seq)
        dataset = NumpyDataset(X=sequences, ids=ids)
        return dataset



import os
import torch
import numpy as np
from typing import List, Sequence, Dict
from deepchem.utils.dft_utils import BaseGrid, RadialGrid


class LebedevLoader(object):
    """load the lebedev points and save the cache to save time

    Examples
    --------
    >>> from deepchem.utils.dft_utils import LebedevLoader
    >>> grid = LebedevLoader.load(3)
    >>> grid.shape
    (6, 3)

    """
    caches: Dict[int, np.ndarray] = {}

    @classmethod
    def load(cls, prec: int) -> np.ndarray:
        """Load the Lebedev grid points with the given precision.

        Parameters
        ----------
        prec: int
            Precision of the Lebedev grid. Must be an odd number between 3 and 131.

        Returns
        -------
        np.ndarray
            Lebedev grid points with shape (nphitheta, 3), where nphitheta
            is the number of points in the grid. The first two columns
            are the polar and azimuthal angles in radians, and the last
            column is the integration weights.

        """
        if prec not in cls.caches:
            # load the lebedev grid points
            dset_path = os.path.join(
                os.path.split(__file__)[0], "lebedevquad",
                "lebedev_%03d.txt" % prec)
            assert os.path.exists(
                dset_path), "The dataset lebedev_%03d.txt does not exist" % prec
            lebedev_dsets = np.loadtxt(dset_path)
            lebedev_dsets[:, :2] *= (np.pi / 180
                                    )  # convert the angles to radians
            # save to the cache
            cls.caches[prec] = lebedev_dsets

        return cls.caches[prec]


class LebedevGrid(BaseGrid):
    """Using Lebedev predefined angular points + radial grid to form 3D grid.

    Lebedev grids. These are specially-constructed grids for quadrature
    on the surface of a sphere,543, 541, 542, 540 based on the octahedral
    point group.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import RadialGrid, LebedevGrid
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> l_grid = LebedevGrid(grid, 3)
    >>> l_grid.get_rgrid().shape
    torch.Size([600, 3])
    >>> grid.get_rgrid().shape
    torch.Size([100, 1])

    """

    def __init__(self, radgrid: RadialGrid, prec: int) -> None:
        """Initialize the Lebedev grid. The grid points are generated by
        combining the radial grid with the Lebedev angular points.

        Parameters
        ----------
        radgrid: RadialGrid
            Radial grid used to generate the Lebedev grid.
        prec: int
            Precision of the Lebedev grid. Must be an odd number between 3 and 131.

        """
        self._dtype = radgrid.dtype
        self._device = radgrid.device

        assert (prec % 2
                == 1) and (3 <= prec <= 131
                          ), "Precision must be an odd number between 3 and 131"

        # load the Lebedev grid points
        lebedev_dsets = torch.tensor(LebedevLoader.load(prec),
                                     dtype=self._dtype,
                                     device=self._device)
        wphitheta = lebedev_dsets[:, -1]  # (nphitheta)
        phi = lebedev_dsets[:, 0]
        theta = lebedev_dsets[:, 1]

        # get the radial grid
        assert radgrid.coord_type == "radial"
        r = radgrid.get_rgrid().unsqueeze(-1)  # (nr, 1)

        # get the cartesian coordinate
        rsintheta = r * torch.sin(theta)
        x = (rsintheta * torch.cos(phi)).view(-1, 1)  # (nr * nphitheta, 1)
        y = (rsintheta * torch.sin(phi)).view(-1, 1)
        z = (r * torch.cos(theta)).view(-1, 1)
        xyz = torch.cat((x, y, z), dim=-1)  # (nr * nphitheta, ndim)
        self._xyz = xyz

        # calculate the dvolume (integration weights)
        dvol_rad = radgrid.get_dvolume().unsqueeze(-1)  # (nr, 1)
        self._dvolume = (dvol_rad * wphitheta).view(-1)  # (nr * nphitheta)

    def get_rgrid(self) -> torch.Tensor:
        """Get the 3D grid points in Cartesian coordinates.

        Returns
        -------
        torch.Tensor
            3D grid points in Cartesian coordinates with shape (ngrid, 3).

        """
        return self._xyz

    def get_dvolume(self) -> torch.Tensor:
        """Get the integration weights for the 3D grid points.

        Returns
        -------
        torch.Tensor
            Integration weights for the 3D grid points with shape (ngrid,).

        """
        return self._dvolume

    @property
    def coord_type(self) -> str:
        """Return the coordinate type of the grid.

        Returns
        -------
        str
            Coordinate type of the grid. Always 'cart'.

        """
        return "cart"

    @property
    def dtype(self) -> torch.dtype:
        """Return the data type of the grid.

        Returns
        -------
        torch.dtype
            Data type of the grid.

        """
        return self._dtype

    @property
    def device(self) -> torch.device:
        """Return the device of the grid.

        Returns
        -------
        torch.device
            Device of the grid.

        """
        return self._device

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """Return the parameter names for serialization.

        Parameters
        ----------
        methodname: str
            Method name.
        prefix: str (default '')
            Prefix to be added to the parameter names.

        Returns
        -------
        List[str]
            List of parameter names.

        """
        if methodname == "get_rgrid":
            return [prefix + "_xyz"]
        elif methodname == "get_dvolume":
            return [prefix + "_dvolume"]
        else:
            raise KeyError("Invalid methodname: %s" % methodname)


class TruncatedLebedevGrid(LebedevGrid):
    """A class to represent the truncated lebedev grid. It is represented
    by various radial grid (usually the sliced ones) with different precisions.

    Examples
    --------
    >>> from deepchem.utils.dft_utils import RadialGrid, TruncatedLebedevGrid
    >>> grid = RadialGrid(100, grid_integrator="chebyshev",
    ...                   grid_transform="logm3")
    >>> l_grid = TruncatedLebedevGrid([grid, grid], [3, 5])
    >>> l_grid.get_rgrid().shape
    torch.Size([2000, 3])
    >>> grid.get_rgrid().shape
    torch.Size([100, 1])

    """

    def __init__(self, radgrids: Sequence[RadialGrid], precs: Sequence[int]):
        """Initialize the truncated Lebedev grid.

        Parameters
        ----------
        radgrids: Sequence[RadialGrid]
            A list of radial grids used to generate the Lebedev grid.
        precs: Sequence[int]
            A list of precisions of the Lebedev grid. Must be odd numbers
            between 3 and 131.

        """
        assert len(radgrids) == len(precs)
        assert len(precs) > 0
        self.lebedevs = [
            LebedevGrid(radgrid, prec)
            for (radgrid, prec) in zip(radgrids, precs)
        ]
        grid0 = self.lebedevs[0]

        # set the variables to be used in the properties
        self._dtype = grid0.dtype
        self._device = grid0.device
        self._xyz = torch.cat(tuple(grid.get_rgrid() for grid in self.lebedevs),
                              dim=0)
        self._dvolume = torch.cat(tuple(
            grid.get_dvolume() for grid in self.lebedevs),
                                  dim=0)



import torch.nn as nn
from torch import Tensor
from deepchem.models.torch_models import TorchModel
import math
import deepchem as dc
import numpy as np
from typing import List, Union, cast
from deepchem.utils.data_utils import load_from_disk, save_to_disk


class InvertedResidual(nn.Module):
    """
    Inverted Residual block used in MobileNetV2 architecture.

    This block uses a combination of pointwise, depthwise, and another pointwise convolution
    with optional residual connections based on input/output channels and stride.

    Parameters
    ----------
    inp : int
        Number of input channels.
    oup : int
        Number of output channels.
    stride : int
        Stride for depthwise convolution. Must be 1 or 2.
    expand_ratio : float
        Expansion ratio for the hidden dimension. If 1, the input is not expanded.

    Returns
    -------
    use_res_connect : bool
        Whether to use the residual connection.
    conv : nn.Sequential
        The core convolutional operations in the block.

    Examples
    --------
    >>> import torch
    >>> x = torch.randn(1, 16, 32, 32)
    >>> block = InvertedResidual(inp=16, oup=16, stride=1, expand_ratio=1)
    >>> out = block(x)
    >>> out.shape
    torch.Size([1, 16, 32, 32])

    """

    def __init__(self, inp: int, oup: int, stride: int,
                 expand_ratio: float) -> None:
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2], "Stride must be 1 or 2."

        hidden_dim = int(inp * expand_ratio)
        self.use_res_connect = self.stride == 1 and inp == oup

        if expand_ratio == 1:
            self.conv = nn.Sequential(
                # dw
                nn.Conv2d(hidden_dim,
                          hidden_dim,
                          3,
                          stride,
                          1,
                          groups=hidden_dim,
                          bias=False),
                nn.BatchNorm2d(hidden_dim),
                nn.ReLU6(inplace=True),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )
        else:
            self.conv = nn.Sequential(
                # pw
                nn.Conv2d(inp, hidden_dim, 1, 1, 0, bias=False),
                nn.BatchNorm2d(hidden_dim),
                nn.ReLU6(inplace=True),
                # dw
                nn.Conv2d(hidden_dim,
                          hidden_dim,
                          3,
                          stride,
                          1,
                          groups=hidden_dim,
                          bias=False),
                nn.BatchNorm2d(hidden_dim),
                nn.ReLU6(inplace=True),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                nn.BatchNorm2d(oup),
            )

    def forward(self, x: Tensor) -> Tensor:
        """Forward pass of the InvertedResidual block.

        Parameters
        ----------
        x : Tensor
            Input tensor of shape (N, C, H, W)

        Returns
        -------
        Tensor
            Output tensor of shape (N, oup, H_out, W_out)
        """
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV2(nn.Module):
    """
    MobileNetV2 architecture with support for arbitrary input channels (e.g., for 6-channel images).

    This version of MobileNetV2 is modified to accept custom input channels and configurable width scaling.
    It uses inverted residual blocks for efficient mobile/edge inference.

    Parameters
    ----------
    n_class : int, default=1000
        Number of output classes (for classification) or output units (for regression).
    input_size : int, default=224
        Spatial resolution of the input image. Must be divisible by 32.
    width_mult : float, default=1.0
        Width multiplier for the entire network to adjust model size.
    in_channels : int, default=6
        Number of input channels for the first convolutional layer.
    """

    def __init__(self,
                 n_class: int = 1000,
                 input_size: int = 224,
                 width_mult: float = 1.0,
                 in_channels: int = 6) -> None:
        super(MobileNetV2, self).__init__()
        block = InvertedResidual
        input_channel = 32
        last_channel = 1280
        interverted_residual_setting = [
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1],
        ]

        # building first layer
        assert input_size % 32 == 0, "Input size must be divisible by 32."
        self.last_channel = self._make_divisible(
            last_channel * width_mult) if width_mult > 1.0 else last_channel
        self.features: Union[List[nn.Module], nn.Sequential] = [
            self._conv_bn(in_channels, input_channel, 2)
        ]  # Modified to support multi-channel input
        # building inverted residual blocks
        for t, c, n, s in interverted_residual_setting:
            output_channel = self._make_divisible(c *
                                                  width_mult) if t > 1 else c
            for i in range(n):
                if i == 0:
                    self.features.append(
                        block(input_channel, output_channel, s, expand_ratio=t))
                else:
                    self.features.append(
                        block(input_channel, output_channel, 1, expand_ratio=t))
                input_channel = output_channel
        # building last several layers
        self.features.append(self._conv_1x1_bn(input_channel,
                                               self.last_channel))
        # make it nn.Sequential
        self.features = nn.Sequential(*self.features)

        # building classifier
        self.classifier = nn.Linear(self.last_channel, n_class)

        self._initialize_weights()

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass for MobileNetV2.

        Parameters
        ----------
        x : Tensor
            Input tensor of shape (N, in_channels, H, W)

        Returns
        -------
        Tensor
            Output tensor of shape (N, n_class)
        """
        features = cast(nn.Sequential, self.features)
        x = features(x)
        x = x.mean(3).mean(2)  # Global average pooling
        x = self.classifier(x)
        return x

    def _initialize_weights(self) -> None:
        """Initialize weights using He initialization for conv and standard init for BN and Linear layers."""
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                n = m.weight.size(1)
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()

    def _conv_bn(self, inp: int, oup: int, stride: int) -> nn.Sequential:
        """3x3 Convolution + BatchNorm + ReLU6"""
        return nn.Sequential(nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
                             nn.BatchNorm2d(oup), nn.ReLU6(inplace=True))

    def _conv_1x1_bn(self, inp: int, oup: int) -> nn.Sequential:
        """1x1 Convolution + BatchNorm + ReLU6"""
        return nn.Sequential(nn.Conv2d(inp, oup, 1, 1, 0, bias=False),
                             nn.BatchNorm2d(oup), nn.ReLU6(inplace=True))

    def _make_divisible(self, x: float, divisible_by: int = 8) -> int:
        """
        Make channel size divisible by `divisible_by`, used for efficient tensor ops.

        Parameters
        ----------
        x : float
            Original value
        divisible_by : int
            Factor to make x divisible by

        Returns
        -------
        int
            Adjusted value
        """
        return int(np.ceil(x * 1. / divisible_by) * divisible_by)


class MobileNetV2Model(TorchModel):
    """
    MobileNetV2 with multi-channel support and classification/regression modes.

    MobileNetV2 is a lightweight and efficient convolutional neural network
    architecture designed for mobile and edge devices. It builds on the success
    of depthwise separable convolutions introduced in MobileNetV1, but introduces
    two key innovations: inverted residual blocks and linear bottlenecks. Unlike
    traditional residual blocks that compress then expand features, MobileNetV2
    first expands the input channels, applies a depthwise convolution (which
    processes each channel independently), and then projects it back down to a
    lower-dimensional space using a pointwise (1x1) convolution. This "inverted"
    structure preserves rich information in the high-dimensional space while
    maintaining efficiency. A linear layer (without non-linearity) at the bottleneck
    helps retain feature information during compression. Residual connections are
    selectively added when the input and output shapes match, which stabilizes
    training and improves accuracy.

    Parameters
    ----------
    n_tasks: int
        Number of tasks (output dimensions)
    in_channels: int, default 6
        Number of input channels
    input_size: int, default 224
        Input image size (must be divisible by 32)
    mode: str, default "classification"
        Either "regression" or "classification"
    n_classes: int, default 2
        Number of classes to predict (only used in classification mode)
    width_mult: float, default 1.0
        Width multiplier for the network

    References
    ----------
    .. Sandler, M., Howard, A., Zhu, M., Zhmoginov, A., & Chen, L. C. (2018). MobileNetV2: Inverted residuals and linear bottlenecks. *arXiv preprint arXiv:1801.04381*.

    """

    def __init__(self,
                 n_tasks: int,
                 in_channels: int = 6,
                 input_size: int = 224,
                 mode: str = "classification",
                 n_classes: int = 2,
                 width_mult: float = 1.0,
                 **kwargs):
        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")

        self.n_tasks = n_tasks
        self.in_channels = in_channels
        self.mode = mode
        self.n_classes = n_classes

        if mode == 'classification':
            output_dim = n_tasks * n_classes
        else:
            output_dim = n_tasks

        model = MobileNetV2(n_class=output_dim,
                            input_size=input_size,
                            width_mult=width_mult,
                            in_channels=in_channels)

        loss: Union[dc.models.losses.L2Loss,
                    dc.models.losses.SparseSoftmaxCrossEntropy]
        if mode == 'classification':
            loss = dc.models.losses.SparseSoftmaxCrossEntropy()
        else:
            loss = dc.models.losses.L2Loss()

        super(MobileNetV2Model, self).__init__(model=model, loss=loss, **kwargs)

    def save(self):
        """Saves model to disk using joblib."""
        save_to_disk(self.model, self.get_model_filename(self.model_dir))

    def reload(self):
        """Loads model from joblib file on disk."""
        self.model = load_from_disk(self.get_model_filename(self.model_dir))
